@@ -374,111 +374,114 @@ static void verify_extract_name(TLS_SESS_STATE *TLScontext, X509 *peercert,
 
 /* add_namechecks - tell OpenSSL what names to check */
 
-static int add_namechecks(TLS_SESS_STATE *TLScontext,
-			          const TLS_CLIENT_START_PROPS *props)
+static void add_namechecks(TLS_SESS_STATE *TLScontext,
+			           const TLS_CLIENT_START_PROPS *props)
 {
     SSL    *ssl = TLScontext->con;
-    const char *nexthop = props->nexthop;
-    const char *hname = props->host;
+    int     namechecks_count = 0;
     int     i;
 
+    /* RFC6125: No part-label 'foo*bar.example.com' wildcards for SMTP */
+    SSL_set_hostflags(ssl, X509_CHECK_FLAG_NO_PARTIAL_WILDCARDS);
+
     for (i = 0; i < props->matchargv->argc; ++i) {
 	const char *name = props->matchargv->argv[i];
 	const char *aname;
-	int     want_nexthop = strcasecmp(name, "nexthop") == 0;
-	int     want_dot_nexthop = strcasecmp(name, "dot-nexthop") == 0;
 	int     match_subdomain = 0;
-	int     ret;
 
-	if (want_nexthop || want_dot_nexthop) {
-
-	    /*
-	     * DNS subjectAltNames are required to be ASCII.
-	     * 
-	     * Per RFC 6125 Section 6.4.4 Matching the CN-ID, follows the same
-	     * rules (6.4.1, 6.4.2 and 6.4.3) that apply to subjectAltNames.
-	     * In particular, 6.4.2 says that the reference identifier is
-	     * coerced to ASCII, but no conversion is stated or implied for
-	     * the CN-ID, so it seems it only matches if it is all ASCII.
-	     * Otherwise, it is some other sort of name.
-	     */
-#ifndef NO_EAI
-	    if (!allascii(nexthop) && (aname = midna_domain_to_ascii(nexthop)) != 0) {
-		if (msg_verbose)
-		    msg_info("%s asciified to %s", nexthop, aname);
-		nexthop = aname;
-	    }
-#endif
-	    if (want_nexthop) {
-		ret = SSL_add1_host(ssl, nexthop);
-	    } else {
-		char   *dot_nexthop = concatenate(".", nexthop, (char *) 0);
-
-		ret = SSL_add1_host(ssl, dot_nexthop);
-		myfree(dot_nexthop);
-	    }
-	    if (ret)
-		continue;
-	    return (0);
-	} else if (strcasecmp(name, "hostname") == 0) {
-	    if (SSL_add1_host(ssl, hname))
-		continue;
-	    return (0);
-	}
-	if (*name == '.') {
-	    if (*++name == 0)
-		return (0);
+	if (strcasecmp(name, "nexthop") == 0) {
+	    name = props->nexthop;
+	} else if (strcasecmp(name, "dot-nexthop") == 0) {
+	    name = props->nexthop;
 	    match_subdomain = 1;
-	}
+	} else if (strcasecmp(name, "hostname") == 0) {
+	    name = props->host;
+	} else {
+	    if (*name == '.') {
+		if (*++name == 0) {
+		    msg_warn("%s: ignoring invalid match name: \".\"",
+			     TLScontext->namaddr);
+		    continue;
+		}
+		match_subdomain = 1;
+	    }
 #ifndef NO_EAI
-	else {
+	    else {
 
-	    /*
-	     * Besides U+002E (full stop) IDNA2003 allows labels to be
-	     * separated by any of the Unicode variants U+3002 (ideographic
-	     * full stop), U+FF0E (fullwidth full stop), and U+FF61
-	     * (halfwidth ideographic full stop). Their respective UTF-8
-	     * encodings are: E38082, EFBC8E and EFBDA1.
-	     * 
-	     * IDNA2008 does not permit (upper) case and other variant
-	     * differences in U-labels. The midna_domain_to_ascii() function,
-	     * based on UTS46, normalizes such differences away.
-	     * 
-	     * The IDNA to_ASCII conversion does not allow empty leading labels,
-	     * so we handle these explicitly here.
-	     */
-	    unsigned char *cp = (unsigned char *) name;
+		/*
+		 * Besides U+002E (full stop) IDNA2003 allows labels to be
+		 * separated by any of the Unicode variants U+3002
+		 * (ideographic full stop), U+FF0E (fullwidth full stop), and
+		 * U+FF61 (halfwidth ideographic full stop). Their respective
+		 * UTF-8 encodings are: E38082, EFBC8E and EFBDA1.
+		 * 
+		 * IDNA2008 does not permit (upper) case and other variant
+		 * differences in U-labels. The midna_domain_to_ascii()
+		 * function, based on UTS46, normalizes such differences
+		 * away.
+		 * 
+		 * The IDNA to_ASCII conversion does not allow empty leading
+		 * labels, so we handle these explicitly here.
+		 */
+		unsigned char *cp = (unsigned char *) name;
 
-	    if ((cp[0] == 0xe3 && cp[1] == 0x80 && cp[2] == 0x82)
-		|| (cp[0] == 0xef && cp[1] == 0xbc && cp[2] == 0x8e)
-		|| (cp[0] == 0xef && cp[1] == 0xbd && cp[2] == 0xa1)) {
-		if (name[3]) {
-		    name = name + 3;
-		    match_subdomain = 1;
+		if ((cp[0] == 0xe3 && cp[1] == 0x80 && cp[2] == 0x82)
+		    || (cp[0] == 0xef && cp[1] == 0xbc && cp[2] == 0x8e)
+		    || (cp[0] == 0xef && cp[1] == 0xbd && cp[2] == 0xa1)) {
+		    if (name[3]) {
+			name = name + 3;
+			match_subdomain = 1;
+		    }
 		}
 	    }
+#endif
 	}
+
+	/*
+	 * DNS subjectAltNames are required to be ASCII.
+	 * 
+	 * Per RFC 6125 Section 6.4.4 Matching the CN-ID, follows the same rules
+	 * (6.4.1, 6.4.2 and 6.4.3) that apply to subjectAltNames.  In
+	 * particular, 6.4.2 says that the reference identifier is coerced to
+	 * ASCII, but no conversion is stated or implied for the CN-ID, so it
+	 * seems it only matches if it is all ASCII.  Otherwise, it is some
+	 * other sort of name.
+	 */
+#ifndef NO_EAI
 	if (!allascii(name) && (aname = midna_domain_to_ascii(name)) != 0) {
 	    if (msg_verbose)
 		msg_info("%s asciified to %s", name, aname);
 	    name = aname;
 	}
 #endif
+
 	if (!match_subdomain) {
 	    if (SSL_add1_host(ssl, name))
-		continue;
-	    return (0);
+		++namechecks_count;
+	    else
+		msg_warn("%s: error loading match name: \"%s\"",
+			 TLScontext->namaddr, name);
 	} else {
-	    char   *dot_nexthop = concatenate(".", nexthop, (char *) 0);
+	    char   *dot_name = concatenate(".", name, (char *) 0);
 
-	    ret = SSL_add1_host(ssl, dot_nexthop);
-	    myfree(dot_nexthop);
-	    if (ret)
-		continue;
-	    return (0);
+	    if (SSL_add1_host(ssl, dot_name))
+		++namechecks_count;
+	    else
+		msg_warn("%s: error loading match name: \"%s\"",
+			 TLScontext->namaddr, dot_name);
+	    myfree(dot_name);
 	}
     }
-    return (1);
+
+    /*
+     * If we failed to add any names, OpenSSL will perform no namechecks, so
+     * we set the "must_fail" bit to avoid verification false-positives.
+     */
+    if (namechecks_count == 0) {
+	msg_warn("%s: could not configure peer name checks",
+		 TLScontext->namaddr);
+	TLScontext->must_fail = 1;
+    }
 }
 
 /* tls_auth_enable - set up TLS authentication */
@@ -541,15 +544,10 @@ static int tls_auth_enable(TLS_SESS_STATE *TLScontext,
 	}
 	/* RFC7672 Section 3.1.1 specifies no name checks for DANE-EE(3) */
 	SSL_dane_set_flags(TLScontext->con, DANE_FLAG_NO_DANE_EE_NAMECHECKS);
-	SSL_set_hostflags(TLScontext->con, X509_CHECK_FLAG_NO_PARTIAL_WILDCARDS);
 
 	/* Per RFC7672 the SNI name is the TLSA base domain */
 	sni = props->dane->base_domain;
-	if (!add_namechecks(TLScontext, props)) {
-	    msg_warn("%s: could not configure peer name checks",
-		     TLScontext->namaddr);
-	    return (0);
-	}
+	add_namechecks(TLScontext, props);
 	break;
 
     case TLS_LEV_FPRINT:
@@ -574,11 +572,7 @@ static int tls_auth_enable(TLS_SESS_STATE *TLScontext,
 		return (0);
 	    }
 	}
-	if (!add_namechecks(TLScontext, props)) {
-	    msg_warn("%s: could not configure peer name checks",
-		     TLScontext->namaddr);
-	    return (0);
-	}
+	add_namechecks(TLScontext, props);
 	break;
     default:
 	break;
@@ -871,6 +865,8 @@ TLS_SESS_STATE *tls_client_start(const TLS_CLIENT_START_PROPS *props)
 {
     int     sts;
     int     protomask;
+    int     min_proto;
+    int     max_proto;
     const char *cipher_list;
     SSL_SESSION *session = 0;
     TLS_SESS_STATE *TLScontext;
@@ -894,7 +890,7 @@ TLS_SESS_STATE *tls_client_start(const TLS_CLIENT_START_PROPS *props)
      * Per-session protocol restrictions must be applied to the SSL connection,
      * as restrictions in the global context cannot be cleared.
      */
-    protomask = tls_protocol_mask(props->protocols);
+    protomask = tls_proto_mask_lims(props->protocols, &min_proto, &max_proto);
     if (protomask == TLS_PROTOCOL_INVALID) {
 	/* tls_protocol_mask() logs no warning. */
 	msg_warn("%s: Invalid TLS protocol list \"%s\": aborting TLS session",
@@ -962,12 +958,24 @@ TLS_SESS_STATE *tls_client_start(const TLS_CLIENT_START_PROPS *props)
 	tls_free_context(TLScontext);
 	return (0);
     }
+#define CARP_VERSION(which) do { \
+        if (which##_proto != 0) \
+            msg_warn("%s: error setting %simum TLS version to: 0x%04x", \
+                     TLScontext->namaddr, #which, which##_proto); \
+        else \
+            msg_warn("%s: error clearing %simum TLS version", \
+                     TLScontext->namaddr, #which); \
+    } while (0)
 
     /*
      * Apply session protocol restrictions.
      */
     if (protomask != 0)
 	SSL_set_options(TLScontext->con, TLS_SSL_OP_PROTOMASK(protomask));
+    if (!SSL_set_min_proto_version(TLScontext->con, min_proto))
+	CARP_VERSION(min);
+    if (!SSL_set_max_proto_version(TLScontext->con, max_proto))
+	CARP_VERSION(max);
 
     /*
      * When applicable, configure DNS-based or synthetic (fingerprint or
@@ -990,15 +998,15 @@ TLS_SESS_STATE *tls_client_start(const TLS_CLIENT_START_PROPS *props)
      */
     if (TLScontext->dane && TLScontext->dane->tlsa) {
 	int     usable = tls_dane_enable(TLScontext);
+	int     must_fail = usable <= 0;
 
-	TLScontext->must_fail = usable <= 0;
 	if (usable == 0) {
 	    switch (TLScontext->level) {
 	    case TLS_LEV_HALF_DANE:
 	    case TLS_LEV_DANE:
 		msg_warn("%s: all TLSA records unusable, fallback to "
 			 "unauthenticated TLS", TLScontext->namaddr);
-		TLScontext->must_fail = 0;
+		must_fail = 0;
 		TLScontext->level = TLS_LEV_ENCRYPT;
 		break;
 
@@ -1014,6 +1022,7 @@ TLS_SESS_STATE *tls_client_start(const TLS_CLIENT_START_PROPS *props)
 		break;
 	    }
 	}
+	TLScontext->must_fail |= must_fail;
     }
 
     /*
@@ -1043,7 +1052,7 @@ TLS_SESS_STATE *tls_client_start(const TLS_CLIENT_START_PROPS *props)
      * parameters and append it to the serverid.
      */
     TLScontext->serverid =
-	tls_serverid_digest(TLScontext, props, protomask, cipher_list);
+	tls_serverid_digest(TLScontext, props, cipher_list);
 
     /*
      * When authenticating the peer, use 80-bit plus OpenSSL security level
@@ -1101,7 +1110,7 @@ TLS_SESS_STATE *tls_client_start(const TLS_CLIENT_START_PROPS *props)
      * created for us, so we can use it for debugging purposes.
      */
     if (log_mask & TLS_LOG_TLSPKTS)
-	BIO_set_callback(SSL_get_rbio(TLScontext->con), tls_bio_dump_cb);
+	tls_set_bio_callback(SSL_get_rbio(TLScontext->con), tls_bio_dump_cb);
 
     /*
      * If we don't trigger the handshake in the library, leave control over
@@ -1152,7 +1161,7 @@ TLS_SESS_STATE *tls_client_post_connect(TLS_SESS_STATE *TLScontext,
 
     /* Turn off packet dump if only dumping the handshake */
     if ((TLScontext->log_mask & TLS_LOG_ALLPKTS) == 0)
-	BIO_set_callback(SSL_get_rbio(TLScontext->con), 0);
+	tls_set_bio_callback(SSL_get_rbio(TLScontext->con), 0);
 
     /*
      * The caller may want to know if this session was reused or if a new
@@ -1166,7 +1175,7 @@ TLS_SESS_STATE *tls_client_post_connect(TLS_SESS_STATE *TLScontext,
      * Do peername verification if requested and extract useful information
      * from the certificate for later use.
      */
-    if ((peercert = SSL_get_peer_certificate(TLScontext->con)) != 0) {
+    if ((peercert = TLS_PEEK_PEER_CERT(TLScontext->con)) != 0) {
 	TLScontext->peer_status |= TLS_CERT_FLAG_PRESENT;
 
 	/*
@@ -1186,7 +1195,6 @@ TLS_SESS_STATE *tls_client_post_connect(TLS_SESS_STATE *TLScontext,
 		     TLScontext->peer_CN, TLScontext->issuer_CN,
 		     TLScontext->peer_cert_fprint,
 		     TLScontext->peer_pkey_fprint);
-	X509_free(peercert);
     } else {
 	TLScontext->issuer_CN = mystrdup("");
 	TLScontext->peer_CN = mystrdup("");
